

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   AutoAppend Proton Plugin 1.4                                  *
    *   Copyright  2004 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2004-10-05   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include "escape.h"


#define IS_HEX_DIGIT( ch ) ( ( ch >= '0' ) && ( ch <= '9' ) ) || (	( ch >= 'A' ) && ( ch <= 'F' ) ) || ( ( ch >= 'a' ) && ( ch <= 'f' ) )
#define IS_OCT_DIGIT( ch ) ( ( ch >= '0' ) && ( ch <= '7' ) )



/******************************************************************************************************
**  decode_hex                                                                                       **
******************************************************************************************************/

bool decode_hex( char * in, UINT & out )
{
	out = 0;

	UINT in_len = strlen( in );
	while( in_len-- )
	{
		char ch = *( in++ );

		if(
			( ch >= '0' ) &&
			( ch <= '9' )
		)
		{
			out +=	( ch - '0' +  0 )	<< ( in_len << 2 );
		}
		else if(
			( ch >= 'A' ) &&
			( ch <= 'F' )
		)
		{
			out +=	( ch - 'A' + 10 )	<< ( in_len << 2 );
		}
		else if(
			( ch >= 'a' ) &&
			( ch <= 'f' )
		)
		{
			out +=	( ch - 'a' + 10 )	<< ( in_len << 2 );
		}
		else
		{
			return false;
		}
	}

	return true;
}


/******************************************************************************************************
**  decode_oct                                                                                       **
******************************************************************************************************/

bool decode_oct( char * in, UINT & out )
{
	out = 0;

	UINT in_len = strlen( in );
	while( in_len-- )
	{
		char ch = *( in++ );

		if(
			( ch >= '0' ) &&
			( ch <= '7' )
		)
		{
			out +=	( ch - '0'  )	<< ( in_len * 3 );
		}
		else
		{
			return false;
		}
	}

	return true;
}


/******************************************************************************************************
**  decode_escape_sequences                                                                          **
******************************************************************************************************/

bool decode_escape_sequences( char * in, char * & out )
{
	/*

	0x005C	\\			Backslash

	0x????	\x????		Hexadecimal	(1-4 digits)
	0x00??	\0??		Octal		(1-2 digits)

	0x0007	\a			Bell/alarm
	0x0008	\b			Backspace
	0x001B	\e			ASCII Escape character
	0x000C	\f			Formfeed
	0x000A	\n			Newline
	0x000D	\r			Carriage return
	0x0009	\t			Horizontal tab
	0x000B	\v			Vertical tab

	*/


	UINT in_len = strlen( in );
	if( in_len == 0 )
	{
		return false;
	}

	out = new char[ in_len + 1 ];
	UINT out_len = 0;
	UINT app_len = 0;

	bool bInsideEscape	= false;
	bool bInsideHex		= false;
	bool bInsideOct		= false;

	UINT uDigits;
	char * beg = in;

	while( *in != '\0' )
	{
		if( bInsideEscape )
		{
			if( bInsideHex || bInsideOct )
			{
				if( bInsideHex )
				{
					if( IS_HEX_DIGIT( *in ) )
					{
						if( ++uDigits == 4 )													// *
						{
							// Maximum length reached -> decode
							bInsideEscape	= false;
							bInsideHex		= false;

							char * hex = new char[ uDigits + 1 ];
							memcpy( hex, in - uDigits + 1, uDigits );			// *
							hex[ uDigits ] = '\0';

							UINT res;
							decode_hex( hex, res );
							out[ out_len ] = ( char )res;
							out_len++;

							delete [] hex;
						}
					}
					else	// !IS_HEX_DIGIT( *in )
					{
						// End reached -> decode
						bInsideEscape	= false;
						bInsideHex		= false;

						if( uDigits > 0 )
						{
							char * hex = new char[ uDigits + 1 ];
							memcpy( hex, in - uDigits, uDigits );
							hex[ uDigits ] = '\0';

							UINT res;
							decode_hex( hex, res );
							out[ out_len ] = ( char )res;
							out_len++;

							delete [] hex;

							in--;																// *
						}
						else
						{
							// Wrong escape sequence -> plain copy
							out[ out_len     ] = *( in - 1 );
							out[ out_len + 1 ] = *( in     );
							out_len += 2;
						}
					}
				}
				else	// bInsideOct
				{
					if( IS_OCT_DIGIT( *in ) )
					{
						if( ++uDigits == 2 )													// *
						{
							// Maximum length reached -> decode
							bInsideEscape	= false;
							bInsideOct		= false;

							char * oct = new char[ uDigits + 1 ];
							memcpy( oct, in - uDigits + 1, uDigits );			// *
							oct[ uDigits ] = '\0';

							UINT res;
							decode_oct( oct, res );
							out[ out_len ] = ( char )res;
							out_len++;

							delete [] oct;
						}
					}
					else	// IS_OCT_DIGIT( *in )
					{
						// End reached -> decode
						bInsideEscape	= false;
						bInsideOct		= false;

						if( uDigits > 0 )
						{
							char * oct = new char[ uDigits + 1 ];
							memcpy( oct, in - uDigits, uDigits );
							oct[ uDigits ] = '\0';

							UINT res;
							decode_oct( oct, res );
							out[ out_len ] = ( char )res;
							out_len++;

							delete [] oct;

							in--;																// *
						}
						else
						{
							// Wrong escape sequence -> plain copy
							out[ out_len     ] = *( in - 1 );
							out[ out_len + 1 ] = *( in     );
							out_len += 2;
						}
					}
				}

				in++;

				if( !bInsideEscape )
				{
					beg = in;
				}
			}
			else	// !( bInsideHex || bInsideOct )
			{
				switch( *in )
				{
					case '\\':
					{
						out[ out_len ] = ( char )0x5C;
						out_len++;

						break;
					}
					case '0':
					{
						bInsideOct = true;
						uDigits = 0;

						break;
					}
					case 'x':
					{
						bInsideHex = true;
						uDigits = 0;

						break;
					}
					case 'a':
					{
						out[ out_len ] = ( char )0x07;
						out_len++;

						break;
					}
					case 'b':
					{
						out[ out_len ] = ( char )0x08;
						out_len++;

						break;
					}
					case 'e':
					{
						out[ out_len ] = ( char )0x1B;
						out_len++;

						break;
					}
					case 'f':
					{
						out[ out_len ] = ( char )0x0C;
						out_len++;

						break;
					}
					case 'n':
					{
						out[ out_len ] = ( char )0x0A;
						out_len++;

						break;
					}
					case 'r':
					{
						out[ out_len ] = ( char )0x0D;
						out_len++;

						break;
					}
					case 't':
					{
						out[ out_len ] = ( char )0x09;
						out_len++;

						break;
					}
					case 'v':
					{
						out[ out_len ] = ( char )0x0B;
						out_len++;
				
						break;
					}
					default:
					{
						// Wrong escape sequence -> plain copy
						out[ out_len     ] = *( in - 1 );
						out[ out_len + 1 ] = *( in     );
						out_len += 2;
					}
				}

				in++;

				if( !( bInsideHex || bInsideOct ) )
				{
					beg = in;
					bInsideEscape = false;
				}
			}
		}
		else	// !bInsideEscape
		{
			if( *in == '\\' )
			{
				bInsideEscape = true;

				// Copy everything before
				app_len = in - beg;
				if( app_len > 0 )
				{
					memcpy( out + out_len, beg, ( in - beg ) );
					out_len += app_len;
				}
			}

			in++;	
		}
	}

	if( bInsideEscape )
	{
		// Hex or octal is not finished -> finish now
		if( bInsideHex )
		{
			char * hex = new char[ uDigits + 1 ];
			memcpy( hex, in - uDigits, uDigits );
			hex[ uDigits ] = '\0';

			UINT res;
			decode_hex( hex, res );
			out[ out_len ] = ( char )res;
			out_len++;

			delete [] hex;
		}
		else	// bInsideOct
		{
			char * oct = new char[ uDigits + 1 ];
			memcpy( oct, in - uDigits, uDigits );
			oct[ uDigits ] = '\0';

			UINT res;
			decode_oct( oct, res );
			out[ out_len ] = ( char )res;
			out_len++;

			delete [] oct;
		}
	}
	else
	{
		// Copy rest
		app_len = in - beg;
		if( app_len > 0 )
		{
			memcpy( out + out_len, beg, ( in - beg ) );
			out_len += app_len;
		}
	}

	out[ out_len ] = '\0';

	return true;
}
